"""
Python driver for Sartorius and Minebea Intec scales.

Distributed under the GNU General Public License v2
"""
from __future__ import annotations

from sartorius.driver import Scale


def command_line(args_list: list | None = None) -> None:
    """Command line tool exposed through package install."""
    import argparse
    import asyncio
    import json

    parser = argparse.ArgumentParser(description="Read scale status.")
    parser.add_argument('address', help="The serial or IP address of the scale.")
    parser.add_argument('-n', '--no-info', action='store_true', help="Exclude "
                        "scale information. Reduces communication overhead.")
    parser.add_argument('-z', '--zero', action='store_true', help="Tares and "
                        "zeroes the scale.")
    args = parser.parse_args(args_list)

    async def get() -> None:
        async with Scale(address=args.address) as scale:
            if args.zero:
                await scale.zero()
            d: dict = await scale.get()  # type: ignore
            if not args.no_info and d.get('on', True):
                d['info'] = await scale.get_info()
            print(json.dumps(d, indent=4))
    asyncio.run(get())


if __name__ == '__main__':
    command_line()



"""Contains mocks for driver objects for offline testing."""

import asyncio
from random import choice, random
from typing import Any
from unittest.mock import MagicMock

from sartorius.util import ScaleInfo, ScaleReading

from .driver import Scale as RealScale


class AsyncClientMock(MagicMock):
    """Magic mock that works with async methods."""

    async def __call__(self, *args, **kwargs):  # type: ignore [no-untyped-def]
        """Convert regular mocks into into an async coroutine."""
        return super().__call__(*args, **kwargs)


class Scale(RealScale):
    """Mocks the Scale driver for offline testing."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Set up connection parameters with default scale port."""
        super().__init__(*args, **kwargs)
        self.info: ScaleInfo = {"model": "SIWADCP-1-",
                                "serial": "37454321",
                                "software": "00-37-09",
                                }

    async def get(self) -> ScaleReading:
        """Get scale reading."""
        await asyncio.sleep(random() * 0.25)
        return {'stable': True,
                'units': choice(['kg', 'lb']),
                'mass': random() * 100.0,
                'measurement': choice(['net', 'gross'])}  # type: ignore[list-item]

    async def get_info(self) -> ScaleInfo:
        """Get scale model, serial, and software version numbers."""
        await asyncio.sleep(random() * 0.1)
        return self.info

    async def zero(self) -> None:
        """Tare and zero the scale."""
        await asyncio.sleep(random() * 0.1)



"""Test the driver responds with correct data."""
from unittest import mock

import pytest

from sartorius import command_line
from sartorius.mock import Scale
from sartorius.util import ScaleInfo

ADDRESS = 'fakeip:49155'


@pytest.fixture
def scale_driver():
    """Confirm the scale correctly initializes."""
    return Scale(ADDRESS)


@pytest.fixture
def expected_response() -> ScaleInfo:
    """Return mocked scale data."""
    return {'model': 'SIWADCP-1-',
            'serial': '37454321',
            'software': '00-37-09',
            }


@mock.patch('sartorius.Scale', Scale)
def test_driver_cli(capsys):
    """Confirm the commandline interface works."""
    command_line([ADDRESS])
    captured = capsys.readouterr()
    assert '"stable": true' in captured.out


@pytest.mark.parametrize(('response', 'expected'),
                         #  KKKKKK+*AAAAAAAA*EEECRLF    mass   units stable measurement
                         [('N     +   0.1234 g  \r\n', (0.1234, 'g', True, 'net')),
                          ('G     +   9.9999 kg \r\n', (9.9999, 'kg', True, 'gross')),
                          ('N     +   5.4321    \r\n', (5.4321, '', False, 'net'))])
def test_parse(scale_driver, response, expected):
    """Test the response parsing code.

    Scale weight is returned according to the SMA communication standard:
    K K K K K K + * A A A A A A A A * E E E CR LF
    K: ID code character
    +: plus or minus
    *: space
    A: Digit or letter
    E: unit symbol
    """
    result = scale_driver._parse(response)

    assert result['mass'] == expected[0]
    assert result['units'] == expected[1]
    assert result['stable'] == expected[2]


def test_parse_errors(scale_driver):
    """Test error handling of response parsing code."""
    response = 'X     +   0.1234 g  \r\n'
    with pytest.raises(ValueError, match=r'This driver only supports net/gross weight.'):
        scale_driver._parse(response)


async def test_get_response(scale_driver, expected_response):
    """Confirm that the driver returns correct values on get_info() calls."""
    assert expected_response == await scale_driver.get_info()


async def test_readme_example(expected_response):
    """Confirm the readme example using an async context manager works."""

    async def get():
        async with Scale(ADDRESS) as scale:
            await scale.zero()             # Zero and tare the scale
            response = await scale.get()       # Get mass, units, stability
            assert response['stable'] is True  # pyright:ignore[reportTypedDictNotRequiredAccess]
            assert expected_response == await scale.get_info()  # Get model, serial, software
    await get()



"""
A Python driver for Sartorius and Minebea Intec ethernet scales.

Distributed under the GNU General Public License v2+
"""
import logging
from typing import Any

from sartorius.util import Client, ScaleInfo, ScaleReading, SerialClient, TcpClient

logger = logging.getLogger('sartorius')


class Scale:
    """Driver for Sartorius and Minebea Intec ethernet scales.

    This implements a version of the Scale Manufacturers Association
    standardized communications protocol.
    """

    def __init__(self, address: str = '', ip: str = '', port: int = 49155,
                 **kwargs: Any) -> None:
        """Set up connection parameters, IP address and port.

        Accepts either an address string (TCP or serial), or combination
        of port & ip (TCP, backwards compatible)
        """
        if ip != '':
            if ":" in ip:
                port = int(ip.split(":")[1])
                ip = ip.split(':')[0]
                address = ip
            address = f'{ip}:{port}'
        self.units: str = ""
        if address.startswith('/dev') or address.startswith('COM'):  # serial
            self.hw: Client = SerialClient(address=address, **kwargs)
        else:
            if ':' not in address:
                address = f'{address}:{port}'
            self.hw = TcpClient(address=address, **kwargs)

    async def __aenter__(self, *args: Any) -> 'Scale':
        """Provide async enter to context manager."""
        return self

    async def __aexit__(self, *args: Any) -> None:
        """Provide async exit to context manager."""
        return

    async def get(self) -> ScaleReading:
        """Get scale reading."""
        response = await self.hw._write_and_read('\x1bP')
        if not response:
            raise OSError("Unable to get reading from scale.")
        return self._parse(response)

    async def get_info(self) -> ScaleInfo:
        """Get scale model, serial, and software version numbers."""
        model = await self.hw._write_and_read('\x1bx1_')
        serial = await self.hw._write_and_read('\x1bx2_')
        software = await self.hw._write_and_read('\x1bx3_')
        if not (model and serial and software):
            raise OSError("Unable to get information from scale.")
        response: ScaleInfo = {
            'model': model.strip(),
            'serial': serial.strip(),
            'software': software.strip(),
        }
        for item in response.values():
            if (' + ' in item or ' kg' in item):  # type: ignore[operator] # pyright: ignore[reportOperatorIssue]
                logger.error(f"Received malformed data: {response}")
                return {}  # type: ignore
        return response

    async def zero(self) -> None:
        """Tare and zero the scale."""
        await self.hw._write_and_read('\x1bT')

    def _parse(self, response: str) -> ScaleReading:
        """Parse a scale response.

        Scale weight is returned according to the SMA communication standard:
            K K K K K K + * A A A A A A A A * E E E CR LF
        K: ID code character
        +: plus or minus
        *: space
        A: Digit or letter
        E: unit symbol

        Errors are similar:
            S t a t * * * * * E r r * * # # * * * * CR LF
        #: Error code number
        The most common is "Stat       OFF", indicating the unit is plugged in
        but the face plate is off.

        One weird behavior is in the units field. This field is empty when the
        scale is unstable (weight shifting) but filled in when the reading is
        stable. This implementation converts that to a "stable" boolean.
        """
        if len(response) != 22:
            logger.error(f"Received malformed data: {response}")
            return {'on': False}
        id = response[:6].strip()
        if id == 'Stat':
            error = response[9:14].strip()
            logger.warning(f'Could not read: {error}')
            return {'on': False}
        elif id == 'N':
            measurement = 'net'
        elif id == 'G':
            measurement = 'gross'
        else:
            raise ValueError("This driver only supports net/gross weight.")
        mass = float(response[6:16].replace(' ', ''))
        units = response[17:20].strip()
        if units:
            self.units = units
            stable = True
        else:
            units = self.units
            stable = False
        return {
            'mass': mass,
            'units': units,
            'stable': stable,
            'measurement': measurement,  # type:ignore[typeddict-item]
        }



"""Base functionality for async communication.

Distributed under the GNU General Public License v2+
"""
from __future__ import annotations

import asyncio
import logging
from abc import ABC, abstractmethod
from typing import Any, Literal, TypedDict

import serial

logger = logging.getLogger('sartorius')

class ScaleReading(TypedDict, total=False):
    """A single reading from the scale."""

    stable: bool
    units: str
    mass: float
    measurement: Literal['net', 'gross']
    on: Literal[False]


class ScaleInfo(TypedDict):
    """Information about the scale."""

    model: str
    serial: str
    software: str


class Client(ABC):
    """Base class for a generic reconnecting client."""

    def __init__(self, timeout: float) -> None:
        """Initialize the client."""
        self.eol = b'\r\n'
        self.open = False
        self.timeout = timeout
        self.max_timeouts = 10
        self.lock: asyncio.Lock | None = None

    def close(self) -> None:
        """Close the connection."""
        self.open = False

    async def _write_and_read(self, command: str) -> str | None:
        """Write a command and read a response.

        As industrial devices are commonly unplugged, this has been expanded to
        handle recovering from disconnects.  A lock is used to queue multiple requests.
        """
        if not self.lock:
            # lock initialized here so the loop exists.
            self.lock = asyncio.Lock()
        async with self.lock:  # lock releases on CancelledError
            await self._handle_connection()
            if self.open:
                try:
                    response = await self._handle_communication(command)
                except asyncio.IncompleteReadError:
                    logger.error('IncompleteReadError.  Are there multiple connections?')
                    return None
            else:
                response = None
        return response

    @abstractmethod
    async def _handle_connection(self) -> None:
        """Automatically maintain the connection."""
        ...

    @abstractmethod
    async def _handle_communication(self, command: str) -> str | None:
        """Manage communication, including timeouts and logging."""
        ...


class SerialClient(Client):
    """Client using a directly-connected RS232 serial device."""

    def __init__(self, address: str, baudrate: int = 9600, timeout: float = .15,
                 bytesize: int = serial.EIGHTBITS,
                 stopbits: float | int = serial.STOPBITS_ONE,
                 parity: str = serial.PARITY_ODD):
        """Initialize serial port."""
        super().__init__(timeout)
        self.address = address
        assert isinstance(self.address, str)
        self.serial_details = {'baudrate': baudrate,
                               'bytesize': bytesize,
                               'stopbits': stopbits,
                               'parity': parity,
                               'timeout': timeout}
        self.ser = serial.Serial(self.address, **self.serial_details)  # type: ignore [arg-type]

    def close(self) -> None:
        """Close the serial connection."""
        if self.ser and self.ser.is_open:
            self.ser.close()

    async def _handle_connection(self) -> None:
        """Handle the serial connection status."""
        self.open = True

    async def _handle_communication(self, command: str) -> str | None:
        """Manage communication, including timeouts and logging."""
        try:
            self.ser.write(command.encode())
            # don't strip() to keep the line length == 22
            response = self.ser.readline().decode()
            self.timeouts = 0
        except (serial.SerialTimeoutException, serial.SerialException):
            self.timeouts += 1
            if self.timeouts == self.max_timeouts:
                print(f'Reading from {self.address} timed out {self.timeouts} times.')
                self.close()
            return None
        return response


class TcpClient(Client):
    """A generic reconnecting asyncio TCP client.

    This base functionality can be used by any industrial control device
    communicating over TCP.
    """

    def __init__(self, address: str, timeout: float = 1.0):
        """Set connection parameters.

        Connection is handled asynchronously, either using `async with` or
        behind the scenes on the first `await` call.
        """
        super().__init__(timeout)
        try:
            self.address, self.port = address.split(':')
        except ValueError as e:
            raise ValueError('address must be hostname:port') from e
        self.reconnecting = False
        self.timeouts = 0
        self.connection: dict[str, Any] = {}

    def close(self) -> None:
        """Close the TCP connection."""
        if self.open:
            self.connection['writer'].close()
        self.open = False

    async def _connect(self) -> None:
        """Asynchronously open a TCP connection with the server."""
        self.close()
        reader, writer = await asyncio.open_connection(self.address, self.port)
        self.connection = {'reader': reader, 'writer': writer}
        self.open = True

    async def _handle_connection(self) -> None:
        """Automatically maintain TCP connection."""
        try:
            if not self.open:
                await asyncio.wait_for(self._connect(), timeout=0.5)
            self.reconnecting = False
        except (asyncio.TimeoutError, OSError):
            if not self.reconnecting:
                logger.error(f'Connecting to {self.address} timed out.')
            self.reconnecting = True

    async def _handle_communication(self, command: str) -> str | None:
        """Manage communication, including timeouts and logging."""
        try:
            self.connection['writer'].write(command.encode() + self.eol)
            future = self.connection['reader'].readuntil(self.eol)
            line = await asyncio.wait_for(future, timeout=0.5)
            result = line.decode()
            self.timeouts = 0
        except (asyncio.TimeoutError, TypeError, OSError):
            self.timeouts += 1
            if self.timeouts == self.max_timeouts:
                logger.error(f'Reading from {self.address} timed out '
                             f'{self.timeouts} times.')
                self.close()
            result = None
        return result
